@@ -1,8 +1,11 @@
 use anyhow::{anyhow, bail, Result};
 use dialoguer::Confirm;
+use libsdbootconf::{
+    entry::{EntryBuilder, Token},
+    SystemdBootConf,
+};
 use regex::Regex;
-use std::{cmp::Ordering, fmt, fs, io::prelude::*, path::PathBuf};
-use systemd_boot_conf::SystemdBootConf;
+use std::{cmp::Ordering, fmt, fs, io::prelude::*, path::PathBuf, rc::Rc};
 
 use super::{safe_copy, Kernel, REL_ENTRY_PATH};
 use crate::{
@@ -91,14 +94,24 @@ impl Kernel for GenericKernel {
         let kernel_path = self.esp_mountpoint.join(REL_DEST_PATH);
 
         println_with_prefix_and_fl!("remove_kernel", kernel = self.to_string());
-        fs::remove_file(kernel_path.join(&self.vmlinuz))?;
-        fs::remove_file(kernel_path.join(&self.initrd)).ok();
+        let vmlinuz = kernel_path.join(&self.vmlinuz);
+        let initrd = kernel_path.join(&self.initrd);
+
+        fs::remove_file(&vmlinuz)
+            .map_err(|x| eprintln!("WARNING: {}: {}", &vmlinuz.display(), x))
+            .ok();
+        fs::remove_file(&initrd)
+            .map_err(|x| eprintln!("WARNING: {}: {}", &initrd.display(), x))
+            .ok();
 
         println_with_prefix_and_fl!("remove_entry", kernel = self.to_string());
-        fs::remove_file(
-            self.esp_mountpoint
-                .join(format!("loader/entries/{}", self.entry)),
-        )?;
+        let entry = self
+            .esp_mountpoint
+            .join(format!("loader/entries/{}", self.entry));
+
+        fs::remove_file(&entry)
+            .map_err(|x| eprintln!("WARNING: {}: {}", &entry.display(), x))
+            .ok();
 
         self.remove_default()?;
 
@@ -142,31 +155,53 @@ impl Kernel for GenericKernel {
 
         let dest_path = self.esp_mountpoint.join(REL_DEST_PATH);
         let rel_dest_path = PathBuf::from(REL_DEST_PATH);
-        let mut file = fs::File::create(&entry_path)?;
-        let mut buffer = Vec::new();
 
-        writeln!(buffer, "title {} ({})", self.distro, self)?;
-        writeln!(
-            buffer,
-            "linux /{}",
-            rel_dest_path.join(&self.vmlinuz).display()
-        )?;
+        // let mut file = fs::File::create(&entry_path)?;
+        // let mut buffer = Vec::new();
+        //
+        // writeln!(buffer, "title {} ({})", self.distro, self)?;
+        // writeln!(
+        //     buffer,
+        //     "linux /{}",
+        //     rel_dest_path.join(&self.vmlinuz).display()
+        // )?;
+        // dest_path
+        //     .join(UCODE)
+        //     .exists()
+        //     .then(|| writeln!(buffer, "initrd /{}{}", REL_DEST_PATH, UCODE))
+        //     .transpose()?;
+        // dest_path
+        //     .join(&self.initrd)
+        //     .exists()
+        //     .then(|| writeln!(buffer, "initrd /{}{}", REL_DEST_PATH, self.initrd))
+        //     .transpose()?;
+        // writeln!(buffer, "options {}", self.bootarg)?;
+        //
+        // file.write_all(&buffer)?;
+
+        let mut entry = EntryBuilder::new(&self.entry)
+            .title(format!("{} ({})", self.distro, self))
+            .linux(rel_dest_path.join(&self.vmlinuz))
+            .build();
+
         dest_path
             .join(UCODE)
             .exists()
-            .then(|| writeln!(buffer, "initrd /{}{}", REL_DEST_PATH, UCODE))
-            .transpose()?;
-        dest_path
-            .join(&self.initrd)
-            .exists()
-            .then(|| writeln!(buffer, "initrd /{}{}", REL_DEST_PATH, self.initrd))
-            .transpose()?;
-        writeln!(buffer, "options {}", self.bootarg)?;
+            .then(|| entry.tokens.push(Token::Initrd(rel_dest_path.join(UCODE))));
+        dest_path.join(&self.initrd).exists().then(|| {
+            entry
+                .tokens
+                .push(Token::Initrd(rel_dest_path.join(&self.initrd)))
+        });
+        entry.tokens.push(Token::Options(self.bootarg.to_owned()));
 
-        file.write_all(&buffer)?;
+        let mut file = fs::File::create(&entry_path)?;
+        let buffer = entry.to_string();
+
+        file.write_all(buffer.as_bytes())?;
 
         // Make sure the file is complete, otherwise possible ENOSPC (No space left on device)
-        if file.metadata()?.len() != buffer.len() as u64 {
+        if file.metadata()?.len() != buffer.as_bytes().len() as u64 {
             // Remove incomplete file
             fs::remove_file(&entry_path)?;
             bail!(fl!("no_space"));
@@ -179,22 +214,22 @@ impl Kernel for GenericKernel {
     fn set_default(&self) -> Result<()> {
         println_with_prefix_and_fl!("set_default", kernel = self.to_string());
 
-        let mut conf = SystemdBootConf::new(&self.esp_mountpoint)?;
+        let mut conf = SystemdBootConf::load(&self.esp_mountpoint.join("loader/"))?;
 
-        conf.loader_conf.default = Some(Box::from(self.entry.as_str()));
-        conf.overwrite_loader_conf()?;
+        conf.config.default = Some(self.entry.to_owned());
+        conf.write_all()?;
 
         Ok(())
     }
 
     // Remove default entry
     fn remove_default(&self) -> Result<()> {
-        let mut conf = SystemdBootConf::new(&self.esp_mountpoint)?;
+        let mut conf = SystemdBootConf::load(&self.esp_mountpoint.join("loader/"))?;
 
-        if conf.loader_conf.default == Some(Box::from(self.entry.as_str())) {
+        if conf.config.default.as_ref() == Some(&self.entry) {
             println_with_prefix_and_fl!("remove_default", kernel = self.to_string());
-            conf.loader_conf.default = None;
-            conf.overwrite_loader_conf()?;
+            conf.config.default = None;
+            conf.write_all()?;
         }
 
         Ok(())
@@ -241,7 +276,7 @@ impl GenericKernel {
     }
 
     /// Generate a sorted vector of kernel filenames
-    pub fn list(config: &Config) -> Result<Vec<Self>> {
+    pub fn list(config: &Config) -> Result<Vec<Rc<Self>>> {
         // read /usr/lib/modules to get kernel filenames
         let mut kernels = Vec::new();
 
@@ -254,7 +289,7 @@ impl GenericKernel {
                 && dirpath.join("modules.builtin").exists()
             {
                 match Self::parse(config, &dirname) {
-                    Ok(k) => kernels.push(k),
+                    Ok(k) => kernels.push(Rc::new(k)),
                     Err(_) => {
                         println_with_prefix_and_fl!("skip_unidentified_kernel", kernel = dirname);
                     }
@@ -272,7 +307,7 @@ impl GenericKernel {
     }
 
     /// Generate installed kernel list
-    pub fn list_installed(config: &Config) -> Result<Vec<Self>> {
+    pub fn list_installed(config: &Config) -> Result<Vec<Rc<Self>>> {
         let mut installed_kernels = Vec::new();
 
         // Construct regex for the template
@@ -292,7 +327,7 @@ impl GenericKernel {
                         .ok_or_else(|| anyhow!(fl!("invalid_kernel_filename")))?
                         .as_str();
 
-                    installed_kernels.push(Self::parse(config, version)?);
+                    installed_kernels.push(Rc::new(Self::parse(config, version)?));
                 }
             }
         }
